
/*
 * File CASCMultiMode.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J8. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J8 only because so is required by the CASC-J8
 * rules, and for no other reason.
 */

#include <fstream>
#include <cstdlib>
#include <csignal>
#include <sstream>

#include "Lib/Portability.hpp"

#if !COMPILER_MSVC

#include "Lib/DHSet.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Exception.hpp"
#include "Lib/Int.hpp"
#include "Lib/StringUtils.hpp"
#include "Lib/System.hpp"
#include "Lib/TimeCounter.hpp"
#include "Lib/Timer.hpp"
#include "Lib/ScopedPtr.hpp"

#include "Lib/Sys/Multiprocessing.hpp"
#include "Lib/Sys/SyncPipe.hpp"

#include "Shell/Options.hpp"
#include "Shell/Normalisation.hpp"
#include "Saturation/ProvingHelper.hpp"
#include "Shell/Statistics.hpp"
#include "Shell/UIHelper.hpp"

#include "Parse/TPTP.hpp"

#include "CASCMode.hpp"
#include "CASCMultiMode.hpp"

#define SLOWNESS 1.05

using namespace CASC;
using namespace std;
using namespace Lib;
using namespace Lib::Sys;
using namespace Saturation;

bool CASCMultiMode::perform()
{
  CALL("CASCMultiMode::perform");

  
  
  env.options->setTimeLimitInSeconds(100000);

  UIHelper::szsOutput = true;
  env.options->setProof(Options::Proof::TPTP);
  env.options->setStatistics(Options::Statistics::NONE);

  CASCMultiMode casc;

  bool resValue;
  try {
      resValue = casc.searchForProof();
  } catch (Exception& exc) {
      cerr << "% Exception at proof search level" << endl;
      exc.cry(cerr);
      System::terminateImmediately(1); 
  }

  env.beginOutput();
  if (resValue) {
    env.out()<<"% Success in time "<<Timer::msToSecondsString(env.timer->elapsedMilliseconds())<<endl;
  }
  else {
    env.out()<<"% Proof not found in time "<<Timer::msToSecondsString(env.timer->elapsedMilliseconds())<<endl;
    if (env.remainingTime()/100>0) {
      env.out()<<"% SZS status GaveUp for "<<env.options->problemName()<<endl;
    }
    else {
      
      
      
      env.out()<<"% SZS status Timeout for "<<env.options->problemName()<<endl;
    }
  }
  if (env.options && env.options->timeStatistics()) {
    TimeCounter::printReport(env.out());
  }
  env.endOutput();

  return resValue;
} 



bool CASCMultiMode::performStrategy(Shell::Property* property)
{
  CALL("CASCMultiMode::performStrategy");
  cout << "% Hi Geoff, go and have some cold beer while I am trying to solve this very hard problem!\n";

  Schedule quick;
  Schedule fallback;

  CASC::CASCMode::getSchedules(*property,quick,fallback);

  int terminationTime = env.remainingTime()/100; 
  if(terminationTime <= 0){ return false; }

  StrategySet usedSlices;
  if (runSchedule(quick,usedSlices,false,terminationTime)) {
    return true;
  }
  terminationTime = env.remainingTime()/100;
  if(terminationTime <= 0){ return false; }
  return runSchedule(fallback,usedSlices,true,terminationTime);
} 

bool CASCMultiMode::searchForProof()
{
  CALL("CASCMultiMode::searchForProof");

  env.timer->makeChildrenIncluded();
  TimeCounter::reinitialize();

  prb = UIHelper::getInputProblem(*env.options); 

  Shell::Property* property = prb->getProperty();

  {
    TimeCounter tc(TC_PREPROCESSING);
    env.statistics->phase=Statistics::NORMALIZATION;
    Normalisation norm;
    norm.normalise(*prb);
  }

  env.statistics->phase=Statistics::UNKNOWN_PHASE;

  
  Timer::setTimeLimitEnforcement(false);

  UIHelper::szsOutput=true;

  return performStrategy(property);
} 

static unsigned milliToDeci(unsigned timeInMiliseconds) {
  return timeInMiliseconds/100;
}

bool CASCMultiMode::runSchedule(Schedule& schedule,StrategySet& used,bool fallback,int terminationTime)
{
  CALL("CASCMultiMode::runSchedule");

  
  

#if __APPLE__ || __CYGWIN__
    unsigned coreNumber = 16; 
#else
    unsigned coreNumber = System::getNumberOfCores();
#endif
    if (coreNumber < 1){
        coreNumber = 1;
    }
    unsigned requested = env.options->multicore();
    int parallelProcesses = min(coreNumber,requested);

  
  if(parallelProcesses == 0){
    if(coreNumber >=8){ coreNumber = coreNumber-2; }
    parallelProcesses = coreNumber;
  }
  cout << "% Hi Geoff, using " << parallelProcesses << " parallel processes to solve your hard problem\n" << endl; 

  int processesLeft = parallelProcesses;
  Schedule::BottomFirstIterator it(schedule);
 
  int slices = schedule.length();
  while (it.hasNext()) {
    while (processesLeft) {
      CASCMultiMode::coutLineOutput() << "Slices left: " << slices-- << endl;
      CASCMultiMode::coutLineOutput() << "Processes available: " << processesLeft << endl << flush;
      ASS_G(processesLeft,0);

      int elapsedTime = env.timer->elapsedMilliseconds();
      if (elapsedTime >= terminationTime) {
	
        goto finish_up;
      }

      vstring sliceCode = it.next();
      vstring chopped;

      
      int sliceTime = SLOWNESS * getSliceTime(sliceCode,chopped);
      if (used.contains(chopped)) {
	
	continue;
      }
      used.insert(chopped);
      int remainingTime = terminationTime - elapsedTime;
      if (sliceTime > remainingTime) {
	sliceTime = remainingTime;
      }

      ASS_GE(sliceTime,0);
      if (milliToDeci((unsigned)sliceTime) == 0) {
        
        

        
        goto finish_up;
      }

      pid_t childId=Multiprocessing::instance()->fork();
      ASS_NEQ(childId,-1);
      if (!childId) {
        
        try {
          runSlice(sliceCode,sliceTime); 
        } catch (Exception& exc) {
          cerr << "% Exception at run slice level" << endl;
          exc.cry(cerr);
          System::terminateImmediately(1); 
        }
        ASSERTION_VIOLATION; 
      }
      Timer::syncClock();
      ASS(childIds.insert(childId));
      CASCMultiMode::coutLineOutput() << "slice pid "<< childId << " slice: " << sliceCode
				 << " time: " << (sliceTime/100)/10.0 << endl << flush;
      processesLeft--;
      if (!it.hasNext()) {
	break;
      }
    }

    CASCMultiMode::coutLineOutput() << "No processes available: " << endl << flush;
    if (processesLeft==0) {
      if(waitForChildAndCheckIfProofFound()){ return true; }
      
      processesLeft++;
    }
  }

  finish_up:

  while (parallelProcesses!=processesLeft) {
    ASS_L(processesLeft, parallelProcesses);
    if(waitForChildAndCheckIfProofFound()){ return true; }
    
    processesLeft++;
    Timer::syncClock();
  }
  return false;
} 

bool CASCMultiMode::waitForChildAndCheckIfProofFound()
{
  CALL("CASCMultiMode::waitForChildAndCheckIfProofFound");
  ASS(!childIds.isEmpty());

  int resValue;
  pid_t finishedChild = Multiprocessing::instance()->waitForChildTermination(resValue);
#if VDEBUG
  ALWAYS(childIds.remove(finishedChild));
#endif
  if (!resValue) {
    
    CASCMultiMode::coutLineOutput() << "terminated slice pid " << finishedChild << " (success)" << endl << flush;
    return true;
  }
  
  CASCMultiMode::coutLineOutput() << "terminated slice pid " << finishedChild << " (fail)" << endl << flush;
  return false;
} 

void CASCMultiMode::runSlice(vstring sliceCode, unsigned timeLimitInMilliseconds)
{
  CALL("CASCMultiMode::runSlice");

  Options opt = *env.options;
  opt.readFromEncodedOptions(sliceCode);
  opt.setTimeLimitInDeciseconds(milliToDeci(timeLimitInMilliseconds));
  int stl = opt.simulatedTimeLimit();
  if (stl) {
    opt.setSimulatedTimeLimit(int(stl * SLOWNESS));
  }
  runSlice(opt);
} 

void CASCMultiMode::runSlice(Options& strategyOpt)
{
  CALL("CASCMultiMode::runSlice(Option&)");

  System::registerForSIGHUPOnParentDeath();
  UIHelper::cascModeChild=true;

  int resultValue=1;
  env.timer->reset();
  env.timer->start();
  TimeCounter::reinitialize();
  Timer::setTimeLimitEnforcement(true);

  Options opt = strategyOpt;
  
  opt.setNormalize(false);
  opt.setForcedOptionValues();
  opt.checkGlobalOptionConstraints();
  *env.options = opt; 

  env.beginOutput();
  CASCMultiMode::lineOutput() << opt.testId() << " on " << opt.problemName() << endl;
  env.endOutput();

  ProvingHelper::runVampire(*prb, opt);

  
    if (env.statistics->terminationReason == Statistics::REFUTATION ||
        env.statistics->terminationReason == Statistics::SATISFIABLE) {
    resultValue=0;
  }

  System::ignoreSIGHUP(); 

    bool outputResult = false;
    if (!resultValue) {
        _syncSemaphore.dec(SEM_LOCK); 
        
        if (!_syncSemaphore.get(SEM_PRINTED)) {
            _syncSemaphore.set(SEM_PRINTED,1);
            outputResult = true;
        }
        else{
            env.beginOutput();
            CASCMultiMode::lineOutput() << " found a (different) proof too late" << endl;
            env.endOutput();
        }
        
        _syncSemaphore.inc(SEM_LOCK); 
    }
    
    if(outputResult){
        env.beginOutput();
        UIHelper::outputResult(env.out());
        env.endOutput();
    }

  exit(resultValue);
} 

unsigned CASCMultiMode::getSliceTime(vstring sliceCode,vstring& chopped)
{
  CALL("CASCMode::getSliceTime");

  unsigned pos = sliceCode.find_last_of('_');
  vstring sliceTimeStr = sliceCode.substr(pos+1);
  chopped.assign(sliceCode.substr(0,pos));
  unsigned sliceTime;
  ALWAYS(Int::stringToUnsignedInt(sliceTimeStr,sliceTime));
  ASS_G(sliceTime,0); 

  unsigned time = sliceTime + 1;
  if (time < 10) {
    time++;
  }
  
  return time * 100;
} 

ostream& CASCMultiMode::lineOutput()
{
  CALL("CASCMultiMode::lineOutput");
  return env.out() << "% (" << getpid() << ',' << (env.timer->elapsedMilliseconds()/100)/10.0 << ") ";
} 

ostream& CASCMultiMode::coutLineOutput()
{
  CALL("CASCMultiMode::lineOutput");
  return cout << "% (" << getpid() << ',' << (env.timer->elapsedMilliseconds()/100)/10.0 << ") ";
} 

#endif 
