
/*
 * File TermTransformer.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J8. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J8 only because so is required by the CASC-J8
 * rules, and for no other reason.
 */


#include "SortHelper.hpp"
#include "Term.hpp"

#include "TermTransformer.hpp"
#include "FormulaTransformer.hpp"

namespace Kernel
{

Term* TermTransformer::transform(Term* term)
{
  CALL("TermTransformer::transform(Term* term)");

  if (term->isSpecial()) {
    return transformSpecial(term);
  }

  Stack<TermList*> toDo(8);
  Stack<Term*> terms(8);
  Stack<bool> modified(8);
  Stack<TermList> args(8);
  ASS(toDo.isEmpty());
  ASS(terms.isEmpty());
  modified.reset();
  args.reset();

  modified.push(false);
  toDo.push(term->args());

  for(;;) {
    TermList* tt=toDo.pop();
    if(tt->isEmpty()) {
      if(terms.isEmpty()) {
	
	
	ASS(toDo.isEmpty());
	break;
      }
      Term* orig=terms.pop();
      if(!modified.pop()) {
        if (!orig->isSpecial()) {
          args.truncate(args.length() - orig->arity());
        }
	args.push(TermList(orig));
	continue;
      }

      if (orig->isSpecial()) {
        args.push(TermList(orig));
      } else {
        
        
        
        TermList *argLst = &args.top() - (orig->arity() - 1);
        args.truncate(args.length() - orig->arity());
        args.push(TermList(Term::create(orig, argLst)));
      }
      modified.setTop(true);
      continue;
    } else {
      toDo.push(tt->next());
    }

    TermList tl=*tt;
    TermList dest;
    if (tl.isTerm() && tl.term()->isSpecial()) {
      Term* td = transformSpecial(tl.term());
      if (td == tl.term()) {
        dest = tl;
      } else {
        dest = TermList(td);
      }
    } else {
      dest = transformSubterm(tl);
    }
    if(tl!=dest) {
      args.push(dest);
      modified.setTop(true);
      continue;
    }
    if(tl.isVar()) {
      args.push(tl);
      continue;
    }
    ASS(tl.isTerm());
    Term* t=tl.term();
    terms.push(t);
    modified.push(false);
    if (t->isSpecial()) {
      TermList aux[1];
      aux[0].makeEmpty();
      toDo.push(aux);
    } else {
      toDo.push(t->args());
    }
  }
  ASS(toDo.isEmpty());
  ASS(terms.isEmpty());
  ASS_EQ(modified.length(),1);
  ASS_EQ(args.length(), term->arity());

  if(!modified.pop()) {
    return term;
  }

  ASS_EQ(args.size(), term->arity());
  
  
  
  TermList* argLst=&args.top() - (term->arity() - 1);

  if (term->isLiteral()) {
    return Literal::create(static_cast<Literal*>(term), argLst);
  } else {
    return Term::create(term, argLst);
  }
}

Literal* TermTransformer::transform(Literal* lit)
{
  CALL("TermTransformer::transform(Literal* lit)");
  Term* t = transform(static_cast<Term*>(lit));
  ASS(t->isLiteral());
  return static_cast<Literal*>(t);
}

Term* TermTransformer::transformSpecial(Term* term)
{
  CALL("TermTransformer::transformSpecial(Term* term)");
  ASS(term->isSpecial());

  Term::SpecialTermData* sd = term->getSpecialData();
  switch (sd->getType()) {
    case Term::SF_ITE: {
      Formula* condition = transform(sd->getCondition());
      TermList thenBranch = transform(*term->nthArgument(0));
      TermList elseBranch = transform(*term->nthArgument(1));

      if ((condition == sd->getCondition()) &&
          (thenBranch == *term->nthArgument(0)) &&
          (elseBranch == *term->nthArgument(1))) {
        return term;
      } else {
        return Term::createITE(condition, thenBranch, elseBranch, sd->getSort());
      }
    }

    case Term::SF_FORMULA: {
      Formula* formula = transform(sd->getFormula());

      if (formula != sd->getFormula()) {
        return term;
      } else {
        return Term::createFormula(formula);
      }
    }

    case Term::SF_LET: {
      TermList binding = transform(sd->getBinding());
      TermList body = transform(*term->nthArgument(0));

      if ((binding == sd->getBinding() && (body == *term->nthArgument(0)))) {
        return term;
      } else {
        return Term::createLet(sd->getFunctor(), sd->getVariables(), binding, body, sd->getSort());
      }
    }

    default:
      ASSERTION_VIOLATION_REP(term->toString());
  }
}

TermList TermTransformer::transform(TermList ts)
{
  CALL("TermTransformer::transform(TermList ts)");

  if (ts.isVar()) {
    return transformSubterm(ts);
  } else {
    Term* transformed = transform(ts.term());
    if (transformed != ts.term()) {
      return TermList(transformed);
    } else {
      return ts;
    }
  }
}

Formula* TermTransformer::transform(Formula* f)
{
  CALL("TermTransformer::transform(Formula* f)");
  static TermTransformingFormulaTransformer ttft(*this);
  return ttft.transform(f);
}

Term* TermTransformerTransformTransformed::transform(Term* term)
{
  CALL("TermTransformerTransformTransformed::transform(Term* term)");
  ASS(term->shared());

  static Stack<TermList*> toDo(8);
  static Stack<Term*> terms(8);
  static Stack<TermList> args(8);
  ASS(toDo.isEmpty());
  ASS(terms.isEmpty());
  args.reset();

  toDo.push(term->args());

  

  for(;;) {
    TermList* tt=toDo.pop();
    if(tt->isEmpty()) {
      

      if(terms.isEmpty()) {
        
        
        ASS(toDo.isEmpty());
        break;
      }
      Term* orig=terms.pop();

      

      TermList* argLst = 0;
      if (orig->arity()) {
        
        
        
        argLst=&args.top() - (orig->arity()-1);
        args.truncate(args.length() - orig->arity());
      }

      

      args.push(transformSubterm(TermList(Term::create(orig,argLst))));
      continue;
    } else {
      toDo.push(tt->next());
    }

    

    TermList tl=*tt;
    if(tl.isVar()) {
      TermList dest=transformSubterm(tl);
      args.push(dest);
      continue;
    }
    ASS(tl.isTerm());
    Term* t=tl.term();
    terms.push(t);
    toDo.push(t->args());
  }
  ASS(toDo.isEmpty());
  ASS(terms.isEmpty());
  ASS_EQ(args.length(), term->arity());

  ASS_EQ(args.size(), term->arity());
  
  
  
  TermList* argLst=&args.top() - (term->arity() - 1);
  if (term->isLiteral()) {
    return Literal::create(static_cast<Literal*>(term), argLst);
  } else {
    return Term::create(term, argLst);
  }
}

Literal* TermTransformerTransformTransformed::transform(Literal* lit)
{
  CALL("TermTransformer::transform(Literal* lit)");
  Term* t = transform(static_cast<Term*>(lit));
  ASS(t->isLiteral());
  return static_cast<Literal*>(t);
}

Term* TermTransformerTransformTransformed::transformSpecial(Term* term)
{
  CALL("TermTransformerTransformTransformed::transformSpecial(Term* term)");
  ASS(term->isSpecial());

  Term::SpecialTermData* sd = term->getSpecialData();
  switch (sd->getType()) {
    case Term::SF_ITE: {
      Formula* condition = transform(sd->getCondition());
      TermList thenBranch = transform(*term->nthArgument(0));
      TermList elseBranch = transform(*term->nthArgument(1));

      if ((condition == sd->getCondition()) &&
          (thenBranch == *term->nthArgument(0)) &&
          (elseBranch == *term->nthArgument(1))) {
        return term;
      } else {
        return Term::createITE(condition, thenBranch, elseBranch, sd->getSort());
      }
    }

    case Term::SF_FORMULA: {
      Formula* formula = transform(sd->getFormula());

      if (formula != sd->getFormula()) {
        return term;
      } else {
        return Term::createFormula(formula);
      }
    }

    case Term::SF_LET: {
      TermList binding = transform(sd->getBinding());
      TermList body = transform(*term->nthArgument(0));

      if ((binding == sd->getBinding() && (body == *term->nthArgument(0)))) {
        return term;
      } else {
        return Term::createLet(sd->getFunctor(), sd->getVariables(), binding, body, sd->getSort());
      }
    }

    default:
      ASSERTION_VIOLATION_REP(term->toString());
  }
}

TermList TermTransformerTransformTransformed::transform(TermList ts)
{
  CALL("TermTransformerTransformTransformed::transform(TermList ts)");

  if (ts.isVar()) {
    return transformSubterm(ts);
  } else {
    Term* transformed = transform(ts.term());
    if (transformed != ts.term()) {
      return TermList(transformed);
    } else {
      return ts;
    }
  }
}

Formula* TermTransformerTransformTransformed::transform(Formula* f)
{
  CALL("TermTransformerTransformTransformed::transform(Formula* f)");
  static TermTransformerTransformTransformedFormulaTransformer ttft(*this);
  return ttft.transform(f);
}

}
